# Python wrapper to have available all the scoring classes in one place and
# merge the results in a single output.
import argparse
import csv
import json
import os
import sys

from .rrm_rna_functions import HMMScanner, LogoGenerator, RNAScoring

__version__ = "1.0.11"


def main():
    print(f"Executing rrmscorer version {__version__}.")

    usr_input_handler = UserInputHandler()
    usr_input = usr_input_handler.parse_args()

    Manager(usr_input).input_handler()


class Manager:
    """
    This class is used to manage the input and output of the scoring framework
    """

    STANDARD_AMINO_ACIDS = {
        "A",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "K",
        "L",
        "M",
        "N",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "V",
        "W",
        "Y",
    }
    STANDARD_RNA_NUCLEOTIDES = {"A", "C", "G", "U"}
    N_MERS = [10, 50, 100, 250]

    # This is the general input manager of the scoring framework
    def __init__(self, usr_input):
        self.usr_input = usr_input
        self._rna_scoring = None
        self._hmm_scan = None
        self._logo_gen = None

    @property
    def rna_scoring(self):
        if not self._rna_scoring:
            self._rna_scoring = RNAScoring()

        return self._rna_scoring

    @property
    def hmm_scan(self):
        if not self._hmm_scan:
            self._hmm_scan = HMMScanner()

        return self._hmm_scan

    @property
    def logo_gen(self):
        if not self._logo_gen:
            self._logo_gen = LogoGenerator()

        return self._logo_gen

    def input_handler(self):
        if self.usr_input.fasta_file:
            seqs_dict = self.hmm_scan.hmmalign_RRMs(
                fasta_file=self.usr_input.fasta_file
            )
        elif self.usr_input.up_id:
            seqs_dict = self.hmm_scan.get_UP_seq(UP_id=self.usr_input.up_id)
        else:
            sys.exit("Invalid input parameters")

        for seq_id, seq in seqs_dict.items():
            self.handle_sequence(seq_id, seq)
            if self.usr_input.aligned:
                os.makedirs(os.path.abspath(self.usr_input.aligned), exist_ok=True)

                aln_out_path = os.path.join(
                    os.path.abspath(self.usr_input.aligned),
                    seq_id.replace("/", "_").replace("|", "_") + "_aligned" + ".fasta",
                )

                with open(aln_out_path, "w", encoding="utf-8") as aln_out:
                    aln_out.write(">{}\n{}\n".format(seq_id, seq))

    def handle_sequence(self, seq_id, seq):
        set_seq = set(seq.upper())
        set_seq.remove("-")
        if set_seq.issubset(Manager.STANDARD_AMINO_ACIDS):
            pass
        else:
            print(
                "\033[91m[ERROR] The protein sequence contains"
                " non-standard amino acids.\033[0m",
                file=sys.stderr,
            )
            sys.exit("The protein sequence contains non-standard amino acids.")
        seq_id = seq_id.replace("/", "_").replace("|", "_")

        print(f"\nRunning predictions for {seq_id}...")
        if self.usr_input.top:
            top_scores = self.rna_scoring.find_best_rna(rrm_seq=seq)

            if self.usr_input.json:
                json_path = os.path.join(
                    os.path.abspath(self.usr_input.json), f"{seq_id}_top_scorers.json"
                )
                os.makedirs(os.path.abspath(self.usr_input.json), exist_ok=True)

                with open(json_path, "w", encoding="utf-8") as fp:
                    json.dump(top_scores, fp, indent=2)
                    print(f"Json file successfully saved in {json_path}")

            if self.usr_input.plot:
                for n_mers in Manager.N_MERS:  # Right line
                    plot_path = os.path.join(
                        os.path.abspath(self.usr_input.plot),
                        f"{seq_id}_top_{n_mers}_logo.png",
                    )
                    os.makedirs(os.path.abspath(self.usr_input.plot), exist_ok=True)

                    self.logo_gen.generate_logo_to_file(
                        plot_path,
                        top_scores,
                        n_mers,
                        self.usr_input.window_size,
                        title=self.usr_input.title,
                        wrap_title=self.usr_input.wrap_title,
                    )
                    print(f"Plot successfully saved in {plot_path}")

        elif self.usr_input.rna_seq:
            if not set(self.usr_input.rna_seq).issubset(
                Manager.STANDARD_RNA_NUCLEOTIDES
            ):
                print(
                    "\033[91m[ERROR] The RNA sequence contains"
                    " non-standard RNA nucleotides.\033[0m"
                )
                sys.exit("Invalid input parameters")

            self.rna_scoring.score_out_seq(
                rrm_seq=seq,
                rna_seq=self.usr_input.rna_seq,
                rna_pos_range=self.usr_input.rna_pos_range,
            )

            for key, score in self.rna_scoring.scores_dict.items():
                adjusted_score = score + 0.89 if self.usr_input.adjust_scores else score
                print(key, adjusted_score)

            if self.usr_input.json:
                json_path = os.path.join(
                    os.path.abspath(self.usr_input.json), f"{seq_id}.json"
                )

                os.makedirs(os.path.abspath(self.usr_input.json), exist_ok=True)

                # Adjust scores if flag is enabled
                scores_to_write = self.rna_scoring.scores_dict
                if self.usr_input.adjust_scores:
                    scores_to_write = {k: v + 0.89 for k, v in self.rna_scoring.scores_dict.items()}

                with open(json_path, "w", encoding="utf-8") as fp:
                    json.dump(scores_to_write, fp, indent=2)
                    print(f"Json file successfully saved in {json_path}")

            if self.usr_input.csv:
                csv_path = os.path.join(
                    os.path.abspath(self.usr_input.csv), f"{seq_id}.csv"
                )

                os.makedirs(os.path.abspath(self.usr_input.csv), exist_ok=True)

                with open(csv_path, "w", encoding="utf-8") as csv_file:
                    writer = csv.writer(csv_file)
                    for key, value in self.rna_scoring.scores_dict.items():
                        adjusted_value = value + 0.89 if self.usr_input.adjust_scores else value
                        writer.writerow([key, adjusted_value])
                    print(f"CSV file successfully saved in {csv_path}")

            if self.usr_input.plot:
                plot_path = os.path.join(
                    os.path.abspath(self.usr_input.plot), f"{seq_id}.png"
                )
                os.makedirs(os.path.abspath(self.usr_input.plot), exist_ok=True)

                self.rna_scoring.plot_rna_kde_to_file(
                    self.usr_input.rna_seq,
                    self.usr_input.window_size,
                    plot_path,
                    x_min=self.usr_input.x_min,
                    x_max=self.usr_input.x_max,
                    title=self.usr_input.title,
                    wrap_title=self.usr_input.wrap_title,
                )
                print(f"Plot(s) successfully saved in {plot_path}")


class UserInputHandler:
    """
    This class is used to handle the user input and parse it into a structured format
    """

    def __init__(self):
        self.parser = self._build_parser()

    def _build_parser(self):
        parser = argparse.ArgumentParser(
            description=f"RRM-RNA scoring version {__version__}"
        )

        input_arg = parser.add_mutually_exclusive_group(required=True)
        input_arg.add_argument(
            "-u", "--uniprot", help="UniProt identifier", metavar="UNIPROT_ID"
        )
        input_arg.add_argument(
            "-f", "--fasta", help="Fasta file path", metavar="/path/to/input.fasta"
        )

        feat_arg = parser.add_mutually_exclusive_group(required=True)
        feat_arg.add_argument(
            "-r", "--rna", help="RNA sequence", metavar="RNA_SEQUENCE"
        )
        feat_arg.add_argument(
            "-t",
            "--top",
            action="store_true",
            help="To find the top scoring RNA fragments",
        )

        parser.add_argument(
            "-w",
            "--window_size",
            required=False,
            help="The window size to test",
            metavar="N",
        )
        parser.add_argument(
            "-j",
            "--json",
            help="Store the results in a json file in the declared directory path",
            metavar="/path/to/output",
        )
        parser.add_argument(
            "-c",
            "--csv",
            help="Store the results in a CSV file in the declared directory path",
            metavar="/path/to/output",
        )
        parser.add_argument(
            "-p",
            "--plot",
            help="Store the plots in the declared directory path",
            metavar="/path/to/output",
        )
        parser.add_argument(
            "-a",
            "--aligned",
            help="Store the aligned sequences in the declared directory path",
            metavar="/path/to/output",
        )
        parser.add_argument(
            "--x_min",
            type=float,
            default=-0.9,
            help="Minimum value for x-axis in plots (default: -0.9)",
        )
        parser.add_argument(
            "--x_max",
            type=float,
            default=1.0,
            help="Maximum value for x-axis in plots (default: 1.0)",
        )
        parser.add_argument(
            "--title",
            type=str,
            help="Title for the generated plots",
        )
        parser.add_argument(
            "--wrap-title",
            action="store_true",
            help="Wrap long titles to multiple lines",
        )
        parser.add_argument(
            "--adjust-scores",
            action="store_true",
            help="Add 0.89 to scores to better separate training and randomized regions (positive scores indicate likely binders, negative scores indicate less likely binders)",
        )
        parser.add_argument(
            "-v",
            "--version",
            action="version",
            help="show RRM-RNA scoring version number and exit",
            version=f"RRM-RNA scoring {__version__}",
        )

        return parser

    def parse_args(self):
        """
        Parse the user input and return it in a structured format
        """

        usr_input = UserInput()

        input_files = self.parser.parse_args()

        usr_input.fasta_file = input_files.fasta
        usr_input.up_id = input_files.uniprot

        usr_input.rna_seq = input_files.rna
        usr_input.top = input_files.top

        # User defined outputs
        usr_input.json = input_files.json
        usr_input.csv = input_files.csv
        usr_input.plot = input_files.plot
        usr_input.aligned = input_files.aligned
        usr_input.x_min = input_files.x_min
        usr_input.x_max = input_files.x_max
        usr_input.title = input_files.title
        usr_input.wrap_title = input_files.wrap_title
        usr_input.adjust_scores = input_files.adjust_scores

        # Default window size
        if input_files.window_size:
            usr_input.window_size = int(input_files.window_size)

            if usr_input.window_size == 3:
                usr_input.rna_pos_range = (3, 6)

            elif usr_input.window_size == 5:
                usr_input.rna_pos_range = (2, 7)

            else:
                sys.exit("Only 3 and 5 nucleotide windows are accepted")
        else:  # Default ws=5 if not in input
            usr_input.window_size = 5
            usr_input.rna_pos_range = (2, 7)

        return usr_input


class UserInput:
    """
    This class is used to store the user input in a structured way
    """

    def __init__(self):
        self.fasta_file = None
        self.up_id = None
        self.rna_seq = None
        self.top = None
        self.json = None
        self.csv = None
        self.plot = None
        self.aligned = None
        self.window_size = None
        self.rna_pos_range = None
        self.x_min = None
        self.x_max = None
        self.title = None
        self.wrap_title = False
        self.adjust_scores = False


if __name__ == "__main__":
    main()



import glob
import itertools
import json
import pathlib
import subprocess
import sys
import tempfile
from collections import Counter
from itertools import product
from os import path

# Set matplotlib backend to non-interactive 'Agg' to prevent GUI session
import matplotlib
matplotlib.use('Agg')

import logomaker
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import requests
from Bio import AlignIO, SeqIO
from matplotlib.cm import ScalarMappable
from sklearn.neighbors import KernelDensity

abs_path = str(pathlib.Path(__file__).parent.resolve())


class HMMScanner:

    def get_UP_seq(self, UP_id):
        URL = "https://rest.uniprot.org/uniprotkb/{}.fasta".format(UP_id)
        response = requests.get(URL)
        with tempfile.NamedTemporaryFile() as self.temp_UP_file:
            open(self.temp_UP_file.name, "wb").write(response.content)
            return self.hmmalign_RRMs(self.temp_UP_file.name)

    def hmmssearch_RRMs(self, fasta_file, outfile):
        bashCmd = [
            "hmmsearch",
            "-A",
            outfile,
            "-E",
            "0.003",
            "--domE",
            "0.003",
            "--incE",
            "0.003",
            abs_path + "/alignment_data/rrm_bound.hmm",  # hmmdb
            "{}".format(fasta_file),  # seqfile
        ]

        process = subprocess.Popen(bashCmd, stdout=subprocess.PIPE)
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            print("command:", bashCmd, file=sys.stderr)
            print("stdout:", stdout, file=sys.stderr)
            print("stdout:", stderr, file=sys.stderr)
            exit("hmmalign failure")

        try:
            align = AlignIO.read(outfile, "stockholm")
            self.n_found_RRM = len(align)
            print("{} RRM domains have been identified".format(self.n_found_RRM))

        except ValueError:  # Error when no sequence is found below threshold
            print("None RRM has been identified in the input sequence/s")
            exit()

    def hmmalign_RRMs(self, fasta_file):
        with tempfile.NamedTemporaryFile() as self.temp_hmmsearch_file, tempfile.NamedTemporaryFile() as self.temp_hmmalign_file:
            # First identify wether there are RRMs and where they are
            self.hmmssearch_RRMs(fasta_file, self.temp_hmmsearch_file.name)

            bashCmd = [
                "hmmalign",
                "--mapali",
                abs_path + "/alignment_data/"
                "rrm_bound_domains_aligned_processed_squeezed.fasta",
                "-o",
                self.temp_hmmalign_file.name,
                abs_path + "/alignment_data/rrm_bound.hmm",
                "{}".format(self.temp_hmmsearch_file.name),
            ]

            process = subprocess.Popen(bashCmd, stdout=subprocess.PIPE)
            stdout, stderr = process.communicate()
            if process.returncode != 0:
                print("command:", bashCmd, file=sys.stderr)
                print("stdout:", stdout, file=sys.stderr)
                print("stdout:", stderr, file=sys.stderr)
                exit("hmmalign failure")

            align = AlignIO.read(self.temp_hmmalign_file.name, "stockholm")
            new_seq_to_map = list(str(align[0].seq).upper())
            seqs_dict = {
                aln_seq.id: str(aln_seq.seq) for aln_seq in align[-self.n_found_RRM :]
            }

            # In case new gaps are added, map back to the original positions
            if len(list(seqs_dict.values())[0]) > 243:
                # Get the new mapping manually #TODO: MUST be a better way...
                original_alignment = SeqIO.parse(
                    open(
                        abs_path
                        + "/alignment_data/rrm_bound_domains_aligned_processed_squeezed.fasta"
                    ),
                    "fasta",
                )

                orig_seq_to_map = list(
                    [str(i.seq).upper() for i in original_alignment][0]
                )

                aln_length = len(orig_seq_to_map)

                new_mapping = {}
                new_gaps = 0
                for i in range(aln_length):
                    if new_seq_to_map[: i + 1] == orig_seq_to_map[: i + 1]:
                        new_mapping[i] = i + new_gaps
                    else:
                        while new_seq_to_map[: i + 1] != orig_seq_to_map[: i + 1]:
                            new_gaps += 1
                            new_seq_to_map.pop(i)
                        new_mapping[i] = i + new_gaps

                # Update the seqs_dict to keep the right size and relevant part
                # IT IS NOT THE REAL SEQUENCE
                for seqid, seq in seqs_dict.items():
                    updated_seq = "".join(
                        [seq[new_mapping[pos]] for pos in new_mapping.keys()]
                    )
                    seqs_dict[seqid] = updated_seq

            return seqs_dict


class RNAScoring:
    def __init__(self):
        self.scores_dict = {}

        # Protein bias
        self.fasta_rna_aln = path.join(
            abs_path, "alignment_data/RNAs_aligned_cluster_0_nt_full_refined.fasta"
        )
        self.fasta_prot_aln = path.join(
            abs_path,
            "alignment_data/rrm_bound_domains_aligned_processed_squeezed.fasta",
        )
        # Input data
        self.master_aln = path.join(
            abs_path,
            "alignment_data/RRM_master_interaction_info_with_nt_unbiased_cluster_0.json",
        )

        self.rna_aln = path.join(
            abs_path, "alignment_data/neighbours_rna_cluster_0.json"
        )


    @property
    def rna_seqs(self):
        return list(SeqIO.parse(open(self.fasta_rna_aln), "fasta"))

    @property
    def rna_seqs_dict_full(self):
        return {i.id: str(i.seq) for i in self.rna_seqs}

    @property
    def bias_dict(self):
        return Counter(
            [i.id.split("_")[0] + "_" + i.id.split("_")[1] for i in self.rna_seqs]
        )

    @property
    def prot_seqs(self):
        return list(SeqIO.parse(open(self.fasta_prot_aln), "fasta"))

    @property
    def prot_seqs_dict_full(self):
        return {"_".join(i.id.split("_")[:2]): str(i.seq) for i in self.prot_seqs}

    @property
    def aln_data(self):
        return json.load(open(self.master_aln))

    @property
    def rna_data(self):
        return json.load(open(self.rna_aln))

    @property
    def entries_list(self):
        return list(self.rna_seqs_dict_full.keys())

    def load_scores(
        self,
        scores_folder=path.join(abs_path, "alignment_data/precalculated_scores/*.pkl"),
    ):
        self.all_df_dict = {}

        for file in glob.glob(scores_folder):
            df = pd.read_pickle(file)
            res_num = int(file.split("/")[-1].split(".")[0].split("_")[0])
            nuc_num = int(file.split("/")[-1].split(".")[0].split("_")[1])

            self.all_df_dict[(res_num, nuc_num)] = df

        return self.all_df_dict

    def score_out_seq(self, rrm_seq, rna_seq, rna_pos_range):
        window_size = rna_pos_range[1] - rna_pos_range[0]

        self.scores_dict = {}
        for i in range(len(rna_seq) - (window_size - 1)):
            rna_window = rna_seq[i : i + window_size]
            self.scores_dict[rna_window] = []
            scores_by_pos = []

            for selected_pos, df in self.load_scores().items():
                if rna_pos_range[0] <= selected_pos[1] < rna_pos_range[1]:
                    nuc = rna_window[selected_pos[1] - rna_pos_range[0]]
                    res = rrm_seq[selected_pos[0]].upper()
                    if res != "-" and nuc != "-":
                        scores_by_pos.append(df.loc[res][nuc])

            self.scores_dict[rna_window] = np.nanmean(scores_by_pos)

        return self.scores_dict

    def plot_rna_kde_to_file(self, rna_seq, window_size, plot_path, dpi=300, x_min=-0.4, x_max=0.8, title=None, wrap_title=False):
        self.plot_rna_kde(
            rna_seq=rna_seq, scores_dict=self.scores_dict, window_size=window_size,
            x_min=x_min, x_max=x_max, title=title, wrap_title=wrap_title
        )
        plt.savefig(plot_path, dpi=dpi)

    def plot_rna_kde(self, rna_seq, scores_dict, window_size, x_min=-0.4, x_max=0.8, title=None, wrap_title=False):
        with open(
            path.join(
                abs_path, "alignment_data/leave-one-out_positive_scores_avg_perc_20.txt"
            ),
            "r",
        ) as pos_file:
            pos_scores = [float(i.strip()) for i in pos_file.readlines()]

        with open(
            path.join(
                abs_path, "alignment_data/leave-one-out_negative_scores_avg_perc_20.txt"
            ),
            "r",
        ) as neg_file:
            neg_scores = [float(i.strip()) for i in neg_file.readlines()]

        pos_kde = KernelDensity(kernel="gaussian", bandwidth=0.5).fit(
            np.array(pos_scores).reshape(-1, 1)
        )
        neg_kde = KernelDensity(kernel="gaussian", bandwidth=0.5).fit(
            np.array(neg_scores).reshape(-1, 1)
        )

        # Match the scores with the RNA fragments (Single RNA fragment)
        labels = []
        score_list = []
        kde_score = []

        for i in range(
            len(rna_seq) - (window_size - 1)
        ):  # All the windows starting positions
            score = (
                scores_dict[rna_seq[i : i + window_size]] + 0.89
            )  # Value from ROC-AUC
            score_list.append(score)
            labels.append(rna_seq[i : i + window_size])
            kde_score.append(
                float(
                    np.exp(pos_kde.score_samples(np.array(score).reshape(-1, 1)))
                    / np.exp(neg_kde.score_samples(np.array(score).reshape(-1, 1)))
                )
            )

        min_kde = float(
            np.exp(pos_kde.score_samples(np.array(-1.7 + 0.89).reshape(-1, 1)))
            / np.exp(neg_kde.score_samples(np.array(-1.7 + 0.89).reshape(-1, 1)))
        )
        max_kde = float(
            np.exp(pos_kde.score_samples(np.array(0 + 0.89).reshape(-1, 1)))
            / np.exp(neg_kde.score_samples(np.array(0 + 0.89).reshape(-1, 1)))
        )

        # Normalize the data
        kde_normalized = [(x - min_kde) / (max_kde - min_kde) for x in kde_score]

        my_cmap = plt.cm.get_cmap("RdYlGn")
        colors = my_cmap(kde_normalized)

        # Adjust figure size based on the number of RNA fragments and title length
        height = max(6, len(score_list) * 0.3)  # Adjust height based on number of fragments
        width = 10  # Default width
        
        # Create figure with adjusted size
        fig, ax = plt.subplots(figsize=(width, height))
        y_pos = np.arange(len(score_list))
        ax.barh(y=y_pos, width=score_list, align="center", color=colors)

        # Fix the colorbar issue by properly setting up the ScalarMappable
        sm = ScalarMappable(cmap=my_cmap, norm=plt.Normalize(0, 1))
        sm.set_array([])  # Set an empty array to avoid warning
        cbar = plt.colorbar(sm, ax=ax, pad=0.15)
        cbar.set_label("Confidence score", rotation=270, labelpad=25)

        ax.yaxis.tick_right()
        ax.yaxis.set_label_position("right")
        ax.set_yticks(y_pos)
        ax.set_yticklabels(labels)
        ax.set_xlim(x_min, x_max)

        plt.gca().invert_yaxis()
        plt.xlabel("Scores")
        
        # Add title if provided with proper formatting
        if title:
            # Adjust font size for long titles
            title_fontsize = 12 if len(title) > 40 else 14
            
            if wrap_title and len(title) > 40:
                # Split title into multiple lines if it's too long
                words = title.split()
                wrapped_title = []
                current_line = []
                
                for word in words:
                    if len(' '.join(current_line + [word])) <= 40:
                        current_line.append(word)
                    else:
                        wrapped_title.append(' '.join(current_line))
                        current_line = [word]
                
                if current_line:
                    wrapped_title.append(' '.join(current_line))
                
                plt.title('\n'.join(wrapped_title), fontsize=title_fontsize)
            else:
                plt.title(title, fontsize=title_fontsize)
            
            # Add more space at the top for the title
            plt.subplots_adjust(top=0.9)

        return fig

    def find_best_rna(self, rrm_seq):
        # To find the best 5-mer RNA for a rrm_seq
        mer5_list = list(product(*LogoGenerator.NT_LIST * 5))
        # Calculate the scores for all the 3-mer RNA fragments
        mer5_scores = {}
        for rna_seq in mer5_list:
            rna_seq = "".join(rna_seq)
            mer5_scores.update(
                self.score_out_seq(rrm_seq, rna_seq=rna_seq, rna_pos_range=(2, 7))
            )

        sort_scores = dict(
            sorted(mer5_scores.items(), key=lambda x: x[1], reverse=True)
        )

        return sort_scores


class LogoGenerator:
    """
    Set of methods to assess the specificity of any single RRM sequence based on
    RRMScorer:
        1. generate_PPM: Takes all the scores for a given RRM and any RNA seq.
        and the number of top scoring nmers (RNA fragments) to use to calculate
        the PPM (Position Probability Matrix) with the observed probabilities of
        each nucleotide in any of the 5 scored positions.

        2. ppm_to_bits: From the ppm extract the associated bits value for each
        nucleotide in any of the 5 scored positions using the classic equation
        for logo generation. This informs on the information contained in each
        position and in each nucleotide specifically.
            (Optional) 3.1 generate_logo: Takes the bits dataframe as an
            argument to make a logo for that particular RRM (.png). The logo
            depends on the number of nmers used in the previous steps.

        3. retrieve_bits_stats: Return both the total information contained in
        the logo, i.e., the sum of all the values in the dataframe; and the
        maximum bits value for a single RNA motif, i.e., the sum of the highest
        bits value in each position.

    """

    NT_LIST = [["A", "C", "G", "U"]]

    def generate_PPM(self, rna_scores, n_mers, ws=5):
        # PPM = Position Probability Matrix
        ppm = pd.DataFrame(columns=["A", "C", "G", "U"])
        for i in range(ws):
            nt_obs = [mer[i] for mer in list(rna_scores.keys())[:n_mers]]
            freqs = [nt_obs.count(nt) / n_mers for nt in LogoGenerator.NT_LIST[0]]
            # Create a new DataFrame with the freqs values
            new_row = pd.DataFrame(
                {"A": [freqs[0]], "C": [freqs[1]], "G": [freqs[2]], "U": [freqs[3]]}
            )
            # Concatenate the new row to the ppm DataFrame
            ppm = pd.concat([ppm, new_row], ignore_index=True)
        return ppm

    def ppm_to_bits(self, ppm):
        bits_df = pd.DataFrame(columns=["A", "C", "G", "U"])
        for ind in ppm.index:
            bits_factor_A = (
                ppm["A"][ind] * np.log2(ppm["A"][ind]) if not ppm["A"][ind] == 0 else 0
            )
            bits_factor_C = (
                ppm["C"][ind] * np.log2(ppm["C"][ind]) if not ppm["C"][ind] == 0 else 0
            )
            bits_factor_G = (
                ppm["G"][ind] * np.log2(ppm["G"][ind]) if not ppm["G"][ind] == 0 else 0
            )
            bits_factor_U = (
                ppm["U"][ind] * np.log2(ppm["U"][ind]) if not ppm["U"][ind] == 0 else 0
            )

            bits_factor = 2 - np.nansum(
                -bits_factor_A - bits_factor_C - bits_factor_G - bits_factor_U
            )

            new_row = pd.DataFrame(
                {
                    "A": [bits_factor * ppm["A"][ind]],
                    "C": [bits_factor * ppm["C"][ind]],
                    "G": [bits_factor * ppm["G"][ind]],
                    "U": [bits_factor * ppm["U"][ind]],
                }
            )

            # Concatenate the new row to the bits_df DataFrame
            bits_df = pd.concat([bits_df, new_row], ignore_index=True)
        return bits_df

    def generate_logo_to_file(
        self, plot_path, top_scores, n_mers, window_size, dpi=300, title=None, wrap_title=False
    ):
        ppm = self.generate_PPM(rna_scores=top_scores, n_mers=n_mers, ws=window_size)
        bits_df = self.ppm_to_bits(ppm=ppm)
        self.generate_logo(bits_df=bits_df, title=title, wrap_title=wrap_title)
        plt.savefig(plot_path, dpi=dpi)

    def generate_logo(self, bits_df, title=None, wrap_title=False):
        bits_df = bits_df.astype("float64")
        
        # Adjust figure size if there's a title
        if title:
            plt.figure(figsize=(10, 6))
            
        # create Logo object
        ww_logo = logomaker.Logo(bits_df, color_scheme="classic")
        # style using Logo methods
        ww_logo.style_xticks(anchor=0, spacing=1)
        # style using Axes methods
        ww_logo.ax.set_ylabel("Information (bits)")
        ww_logo.ax.set_xlim([-1, len(bits_df)])
        
        # Add title if provided with proper formatting
        if title:
            # Adjust font size for long titles
            title_fontsize = 12 if len(title) > 40 else 14
            
            if wrap_title and len(title) > 40:
                # Split title into multiple lines if it's too long
                words = title.split()
                wrapped_title = []
                current_line = []
                
                for word in words:
                    if len(' '.join(current_line + [word])) <= 40:
                        current_line.append(word)
                    else:
                        wrapped_title.append(' '.join(current_line))
                        current_line = [word]
                
                if current_line:
                    wrapped_title.append(' '.join(current_line))
                
                plt.title('\n'.join(wrapped_title), fontsize=title_fontsize)
            else:
                plt.title(title, fontsize=title_fontsize)
            
            # Add more space at the top for the title
            plt.subplots_adjust(top=0.9)

        return ww_logo



class MatrixPlots:
    def __init__(self):
        self.pkl_dir = "precalculated_scores/pkl_files/"
        self.all_df_dict = self.load_scores(self.pkl_dir)

    def load_scores(self, dir_path):
        df_dict = {}
        for file in glob.glob(dir_path + "*.pkl"):
            df = pd.read_pickle(file)
            res_num = int(file.split("/")[-1].split(".")[0].split("_")[0])
            nuc_num = int(file.split("/")[-1].split(".")[0].split("_")[1])
            df_dict[(res_num, nuc_num)] = df
        return df_dict

    def highlight_cell(self, x, y, ax=None, **kwargs):
        self.rect = plt.Rectangle(
            (x - 0.5, y - 0.5), 1, 1, fill=False, **kwargs, clip_on=False
        )
        ax = ax or plt.gca()
        ax.add_patch(self.rect)
        return self.rect

    def plot_all_dfs(self):
        # Iterate over files in the directory
        for res_nuc_pos, df in self.all_df_dict.items():
            print(res_nuc_pos)
            # Generate the matshow plots for the individual positions
            if df.empty == False:
                df = df.dropna(axis=0, thresh=1)
                df = df.round(decimals=2)
                pos, rna_pos = res_nuc_pos[0], res_nuc_pos[1]

                _max = 3
                _min = -3
                fig = plt.figure()
                ax1 = plt.subplot()
                norm = mcolors.TwoSlopeNorm(vmin=_min, vmax=_max, vcenter=0)

                df = df.astype("float").round(decimals=2)

                ax1.matshow(df, cmap=plt.cm.RdBu, aspect="auto", norm=norm)

                for i, x in zip(df.index, range(df.shape[0] + 1)):
                    for j, y in zip(df.columns, range(df.shape[1])):
                        c = df.loc[i][j]
                        if abs(c) > 0.7 * _max:
                            ax1.text(
                                y,
                                x,
                                str(c),
                                va="center",
                                ha="center",
                                color="white",
                                fontsize=12,
                            )
                        else:
                            ax1.text(
                                y,
                                x,
                                str(c),
                                va="center",
                                ha="center",
                                color="black",
                                fontsize=12,
                            )

                # Df general data
                xaxis = np.arange(df.shape[1])
                yaxis = np.arange(df.shape[0])
                ax1.set_xticks(xaxis)
                ax1.set_yticks(yaxis)
                ax1.set_ylabel("Neighbour residues")
                ax1.set_xlabel("RNA nucleotide")
                ax1.set_xticklabels(df.columns, fontsize=12)
                ax1.set_yticklabels(df.index, fontsize=12)
                plt.title("Prot -> {} - RNA -> {}".format(pos, rna_pos), fontsize=14)

                all_combinations = list(itertools.product(df.index, df.columns))
                for res, nuc in all_combinations:
                    if nuc == "No_contact":
                        continue
                    # Get the row index based on the row name
                    row_index = df.index.get_loc(res)
                    # Get the column index based on the column name
                    column_index = df.columns.get_loc(nuc)
                    self.highlight_cell(
                        column_index, row_index, color="yellow", linewidth=3, zorder=10
                    )
                    plt.savefig(
                        "precalculated_scores/png_files/"
                        "{}_{}_highlighted_{}_{}.png".format(pos, rna_pos, res, nuc)
                    )
                    self.rect.set_visible(False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
from pathlib import Path

from setuptools import find_packages, setup


def read_file(file_path: Path) -> str:
    """Utility function to read a file."""
    try:
        return file_path.read_text(encoding="utf-8")
    except FileNotFoundError as e:
        raise FileNotFoundError(f"Unable to find the file: {file_path}") from e


def get_version(file_path: Path) -> str:
    """Extract the package version from the specified file."""
    content = read_file(file_path)
    version_match = re.search(r'^__version__\s*=\s*"(.*?)"', content, re.MULTILINE)
    if version_match:
        return version_match.group(1)
    raise ValueError(f"Unable to find version string in {file_path}")


PACKAGE_NAME = "rrmscorer"
BASE_PATH = Path(__file__).parent

setup(
    name=PACKAGE_NAME,
    version=get_version(BASE_PATH / PACKAGE_NAME / "__main__.py"),
    author="Joel Roca-Martinez (ORCID: 0000-0002-4313-3845), Wim Vranken (ORCID: 0000-0001-7470-4324)",
    author_email="bio2byte@vub.be, wim.vranken@vub.be, joel.roca@ucl.ac.uk",
    maintainer="Joel Roca-Martinez (ORCID: 0000-0002-4313-3845), Wim Vranken (ORCID: 0000-0001-7470-4324), Adrián Díaz (ORCID: 0000-0003-0165-1318)",
    maintainer_email="bio2byte@vub.be, wim.vranken@vub.be, adrian.diaz@vub.be",
    description="RRM-RNA score predictor",
    license="MIT",
    long_description=read_file(BASE_PATH / "README_pip.md"),
    long_description_content_type="text/markdown",
    url="https://bio2byte.be/rrmscorer",
    project_urls={
        "Documentation": "https://bitbucket.org/bio2byte/rrmscorer/raw/master/readme.md",
        "Source": "https://bitbucket.org/bio2byte/rrmscorer/",
        "Say Thanks!": "https://bio2byte.be/info/",
    },
    packages=find_packages(include=[PACKAGE_NAME]),
    include_package_data=True,
    package_data={
        "rrmscorer": [
            "data/*.json",
            "alignment_data/*",
            "alignment_data/precalculated_scores/*",
            "output_files/*",
        ],
    },
    zip_safe=False,
    platforms=["Linux", "MacOS"],
    keywords="proteins RRM RNA predictor sequence bio2byte",
    classifiers=[
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Development Status :: 5 - Production/Stable",
    ],
    python_requires=">=3.9",
    install_requires=[
        "biopython",
        "numpy",
        "pandas",
        "requests",
        "scikit-learn",
        "matplotlib",
        "logomaker",
        "seaborn",
    ],
    extras_require={
        "dev": ["unittest"],
        "docs": [],
    },
    entry_points={
        "console_scripts": [
            "rrmscorer=rrmscorer.__main__:main",
        ]
    },
)
